@@ -1,16 +1,22 @@
+from pathlib import Path
 import logging
 import os
+import string
 import uuid
+from typing import Iterable, Union
 
 import geojson
 import matplotlib.pyplot as plt
 import numpy as np
 import shapely.geometry as shg
 import shapely.wkt as shwkt
+from box import Box
 from osgeo import gdal, gdal_array, ogr, osr, gdalconst
 
 from . import logger
 from . import pathify
+from .common.utils import run_cmd
+from .pathify import get_name
 
 gdal.UseExceptions()
 
@@ -24,8 +30,9 @@ geom_from_ds = lambda ds: np.abs(np.array(ds.GetGeoTransform()).reshape(2, -1))
 resol_from_ds = lambda ds: geom_from_ds(ds)[:, 1:].sum(axis=0)
 ul_from_ds = lambda ds: geom_from_ds(ds)[:, 0]
 
-get_epsg_from = lambda ds: int(
-    gdal.Info(ds, format='json')['coordinateSystem']['wkt'].rsplit('"EPSG","', 1)[-1].split('"')[0])
+
+# get_epsg_from = lambda ds: int(
+#     gdal.Info(ds, format='json')['coordinateSystem']['wkt'].rsplit('"EPSG","', 1)[-1].split('"')[0])
 
 
 def plot_ds(ds, show=False):
@@ -106,7 +113,7 @@ class DataSet:
 
     @property
     def epsg(self):
-        return get_epsg_from(self.ds)
+        return osr.SpatialReference(wkt=self.projection).GetAttrValue('AUTHORITY', 1)
 
     def plot(self, show=False):
         plot_ds(self.ds, show=show)
@@ -131,7 +138,7 @@ class DataSet:
         osng = osr.SpatialReference()
         osng.ImportFromEPSG(to_epsg)
         wgs84 = osr.SpatialReference()
-        wgs84.ImportFromEPSG(get_epsg_from(self.ds))
+        wgs84.ImportFromEPSG(self.epsg)
         tx = osr.CoordinateTransformation(wgs84, osng)
         # Up to here, all  the projection have been defined, as well as a
         # transformation from the from to the  to :)
@@ -210,10 +217,24 @@ class DataSet:
     def geo_info(self):
         return self.transform, self.projection
 
+    @property
+    def raster_count(self):
+        return self.ds.RasterCount
+
     @property
     def raster_sizes(self):
         return self.ds.RasterXSize, self.ds.RasterYSize
 
+    @property
+    def file_meta(self):
+        return Box(self.ds.GetMetadata())
+
+    def get_band_metas(self):
+        band_metas = list()
+        for i in range(self.raster_count):
+            band_metas.append(Box(self.ds.GetRasterBand(i + 1).GetMetadata()))
+        return band_metas
+
     @classmethod
     def from_array(cls, array, geo_info):
         if isinstance(geo_info, str): geo_info = DataSet(geo_info)
@@ -260,7 +281,36 @@ class DataSet:
         return filepath
 
     @logger.trace(skimpy=True)
-    def to_file(self, filepath, driver_name, no_data_value=-9999, dtype=gdal.GDT_Float32):
+    def to_file(self, filepath, driver_name, no_data_value=-9999, dtype=gdal.GDT_Float32,
+                file_meta: dict = None, band_metas: list = None):
+        driver = gdal.GetDriverByName(driver_name)
+        cols, rows = self.ds.GetRasterBand(1).ReadAsArray().shape
+        outdata = driver.Create(filepath, rows, cols, self.raster_count, dtype)
+        outdata.SetGeoTransform(self.transform)  ##sets same geotransform as input
+        outdata.SetProjection(self.projection)  ##sets same projection as input
+        if file_meta:
+            outdata.SetMetadata({str(k): str(v) for k, v in file_meta.items()})
+
+        for band_i in range(self.ds.RasterCount):
+            band_num = band_i + 1
+            arr = self.ds.GetRasterBand(band_num).ReadAsArray()
+            if no_data_value is False:
+                the_array = arr
+            else:
+                the_array = np.where(np.isnan(arr), no_data_value, arr)
+            band = outdata.GetRasterBand(band_num)
+            band.WriteArray(the_array)
+            if band_metas:
+                band.SetMetadata({str(k): str(v) for k, v in band_metas[band_i].items()})
+            if no_data_value is not False:
+                band.SetNoDataValue(no_data_value)  ##if you want these values transparent
+        outdata.FlushCache()  ##saves to disk!!
+        outdata = None
+        band = None
+        return filepath
+
+    @logger.trace(skimpy=True)
+    def to_file_old(self, filepath, driver_name, no_data_value=-9999, dtype=gdal.GDT_Float32):
         band = self.ds.GetRasterBand(1)
         arr = band.ReadAsArray()
         [cols, rows] = arr.shape
@@ -284,6 +334,10 @@ class DataSet:
         self.ds = None
         return out
 
+    def get_info(self, **kwargs):
+        kwargs = dict(dict(format='json'), **kwargs)
+        return gdal.Info(self.ds, options=gdal.InfoOptions(**kwargs))
+
     @logger.trace(skimpy=True)
     def translate(self, options=None, format=None,
                   outputType=gdalconst.GDT_Unknown, bandList=None, maskBand=None,
@@ -317,11 +371,11 @@ class DataSet:
 
     @property
     def wkt(self):
-        return get_geom_from_ds(self.ds).ExportToWkt()
+        return self.geom.wkt
 
     @property
     def geom(self):
-        return shwkt.loads(self.wkt)
+        return shg.shape(self.get_info()['wgs84Extent'])
 
 
 @logger.trace(level=logging.DEBUG)
@@ -442,13 +496,209 @@ def get_geo_extent(ds):
     return geo_ext
 
 
-def get_geom_from_ds(ds):
-    extent = get_geo_extent(ds)
-    ring = ogr.Geometry(ogr.wkbLinearRing)
-    image_polygon = ogr.Geometry(ogr.wkbPolygon)
-    for coors in extent:
-        ring.AddPoint(*coors)
-    ring.AddPoint(*ring.GetPoint())
-    # print(ring.GetPoint()[:-1])
-    image_polygon.AddGeometry(ring)
-    return image_polygon
+class GdalMan(object):
+    def __init__(self, debug=False, lazy=False, **kwargs):
+        self.out_filepath = None
+        self.kwargs = kwargs
+        self.debug = debug
+        self.lazy = lazy
+        self.data = dict()
+
+    def _get_laziness(self, lazy, out_filepath):
+        # if Path(out_filepath).suffix.lower() == '.vrt' and lazy:
+        #     logger.warning('Lazy is set to True, but provided out_filepath is virtual.',
+        #                    dict(provided_vrt_path=out_filepath))
+        #     lazy = False
+        return self.lazy if lazy is None else lazy
+
+    @property
+    def path(self):
+        return self.out_filepath
+
+    @property
+    def ds(self):
+        return DataSet(self.out_filepath)
+
+    def _run_gdal_cmd(self, gdal_func, *args, is_gdal_calc=False, **input_params):
+        params = list()
+        if is_gdal_calc:
+            default_kwargs = self.kwargs.copy()
+            if 'q' in self.kwargs:
+                q = default_kwargs.pop('q')
+                default_kwargs['--quiet'] = q
+            options = dict(default_kwargs, **input_params)
+        else:
+            options = dict(self.kwargs, **input_params)
+        for k, v in options.items():
+            if v is True:
+                if k.startswith('-'):
+                    params.append(k)
+                else:
+                    params.append(f'-{k}')
+            elif not v:
+                continue
+            else:
+                if k.startswith('--'):
+                    params.append(f"{k}={v}")
+                elif k.startswith('-'):
+                    params.append(f"{k} {v}")
+                else:
+                    params.append(f"-{k} {v}")
+        run_cmd(f" ".join(map(str, [gdal_func, *args, *params])), debug=self.debug)
+
+    def _finalize_result(self, out_filepath, label_as, debug=None):
+        self.out_filepath = self.data[label_as or 'path'] = out_filepath
+        debug = debug or self.debug
+        if debug:
+            if isinstance(debug, str):
+                log_level = debug.lower()
+            else:
+                log_level = 'debug'
+            getattr(logger.a(), log_level)("%s: %s", label_as or get_name(self.out_filepath), self.out_filepath).s()
+        if not Path(self.out_filepath).exists():
+            raise FileNotFoundError(self.out_filepath)
+        return self
+
+    def __getitem__(self, item):
+        return self.data[item]
+
+    def __setitem__(self, key, value):
+        self.data[key] = value
+
+    def gdalwarp(self, *srcfiles, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
+        lazy = self._get_laziness(lazy, out_filepath)
+        if not (lazy and os.path.exists(out_filepath)):
+            # if not Path(out_filepath).exists(): os.remove(out_filepath)
+            self._run_gdal_cmd('gdalwarp', *srcfiles, out_filepath, **dict(dict(overwrite=True), **kwargs))
+        return self._finalize_result(out_filepath, label_as, debug)
+
+    def gdalbuildvrt(self, *gdalfiles, out_filepath, label_as=None, debug=None, **kwargs):
+        self._run_gdal_cmd('gdalbuildvrt', out_filepath, *gdalfiles, **dict(dict(overwrite=True), **kwargs))
+        return self._finalize_result(out_filepath, label_as, debug)
+
+    def gdal_translate(self, src_dataset, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
+        lazy = self._get_laziness(lazy, out_filepath)
+        if not (lazy and os.path.exists(out_filepath)):
+            self._run_gdal_cmd('gdal_translate', src_dataset, out_filepath, **kwargs)
+        return self._finalize_result(out_filepath, label_as, debug)
+
+    def gdal_merge(self, *input_files, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
+        lazy = self._get_laziness(lazy, out_filepath)
+        if not (lazy and os.path.exists(out_filepath)):
+            self._run_gdal_cmd('gdal_merge.py', *input_files, **dict(o=out_filepath, **kwargs))
+        return self._finalize_result(out_filepath, label_as, debug)
+
+    def gdaldem(self, mode, input_dem, out_filepath, color_text_file=None,
+                lazy=None, label_as=None, debug=None, **kwargs):
+        if mode == 'color-relief': assert color_text_file is not None
+        lazy = self._get_laziness(lazy, out_filepath)
+        if not (lazy and os.path.exists(out_filepath)):
+            if color_text_file:
+                self._run_gdal_cmd('gdaldem', mode, input_dem, color_text_file, out_filepath, **kwargs)
+            else:
+                self._run_gdal_cmd('gdaldem', mode, input_dem, out_filepath, **kwargs)
+        return self._finalize_result(out_filepath, label_as, debug)
+
+    def gdal_calc(self, untagged_expression, out_filepath, equation_info,
+                  lazy=None, label_as=None, debug=None, **kwargs):
+        lazy = self._get_laziness(lazy, out_filepath)
+        if not (lazy and os.path.exists(out_filepath)):
+            tags = list(string.ascii_uppercase[:len(equation_info)])
+            kwargs = {k.lstrip('-'): v for k, v in kwargs.items()}
+            additional_eqn_info = {k: v for k, v in kwargs.items() if k[0].isupper()}
+
+            calc_info = {item_name: dict(tag=tag, path=path) for (item_name, path), tag in
+                         zip(dict(equation_info, **additional_eqn_info).items(), tags)}
+
+            tagged_expression = untagged_expression.format(
+                **{band_name: v['tag'] for band_name, v in calc_info.items()})
+            options = {v['tag']: v['path'] for v in calc_info.values()}
+
+            undashed_options = {'outfile': out_filepath, 'calc': f'"{tagged_expression}"', 'overwrite': True}
+
+            undashed_options.update(kwargs)
+            if 'q' in undashed_options:
+                q = undashed_options.pop('q')
+                undashed_options['quiet'] = q
+            for k, v in undashed_options.items():
+                options["--" + k] = v
+            self._run_gdal_cmd('gdal_calc.py', is_gdal_calc=True, **options)
+        return self._finalize_result(out_filepath, label_as, debug)
+
+    def ogr2ogr(self, input_datasource, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
+        lazy = self._get_laziness(lazy, out_filepath)
+        if not (lazy and os.path.exists(out_filepath)):
+            self._run_gdal_cmd('ogr2ogr', out_filepath, input_datasource, **dict(dict(overwrite=True), **kwargs))
+        return self._finalize_result(out_filepath, label_as, debug)
+
+    def gdal_contour(self, src_filename, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
+        lazy = self._get_laziness(lazy, out_filepath)
+        if not (lazy and os.path.exists(out_filepath)):
+            self._run_gdal_cmd('gdal_contour', src_filename, out_filepath, **kwargs)
+        return self._finalize_result(out_filepath, label_as, debug)
+
+    def gdal_polygonize(self, raster_file, out_filepath, layer='polygons',
+                        connectedness=4, nomask=False, mask=None, band=None, ogr_format=None, fieldname='DN',
+                        lazy=None, label_as=None, debug=None, q=None):
+        """
+        gdal_polygonize.py [-8] [-nomask] [-mask filename] <raster_file> [-b band]
+                   [-q] [-f ogr_format] <out_file> [layer] [fieldname]
+        :return: self
+        """
+        args_1 = list()
+        if connectedness != 4: args_1.append(f"-{connectedness}")
+        if nomask: args_1.append(f"-nomask")
+        if mask is not None: args_1.append(f"-mask {mask}")
+        args_2 = list()
+        if band is not None: args_2.append(f"-b {band}")
+        if q is not None: args_2.append("-q")
+        if ogr_format is not None: args_2.append(f"-f {ogr_format}")
+        args_3 = list()
+        if layer: args_3.append(layer)
+        if fieldname: args_3.append(fieldname)
+
+        lazy = self._get_laziness(lazy, out_filepath)
+        if not (lazy and os.path.exists(out_filepath)):
+            self._run_gdal_cmd('gdal_polygonize.py', *args_2, raster_file, *args_2, out_filepath, *args_3)
+        return self._finalize_result(out_filepath, label_as, debug)
+
+    def gdal_rasterize(self, src_filename, out_filepath, lazy=None, label_as=None, debug=None, **kwargs):
+        """
+        gdal_rasterize [-b band]* [-i] [-at]
+    {[-burn value]* | [-a attribute_name] | [-3d]} [-add]
+    [-l layername]* [-where expression] [-sql select_statement]
+    [-dialect dialect] [-of format] [-a_srs srs_def] [-to NAME=VALUE]*
+    [-co "NAME=VALUE"]* [-a_nodata value] [-init value]*
+    [-te xmin ymin xmax ymax] [-tr xres yres] [-tap] [-ts width height]
+    [-ot {Byte/Int16/UInt16/UInt32/Int32/Float32/Float64/
+            CInt16/CInt32/CFloat32/CFloat64}]
+    [-optim {[AUTO]/VECTOR/RASTER}] [-q]
+    <src_datasource> <dst_filename>
+        """
+        lazy = self._get_laziness(lazy, out_filepath)
+        if not (lazy and os.path.exists(out_filepath)):
+            self._run_gdal_cmd('gdal_rasterize', src_filename, out_filepath, **kwargs)
+        return self._finalize_result(out_filepath, label_as, debug)
+
+
+def get_geo_info(upper_left_xy: Union[list, tuple, np.ndarray],
+                 resol_xy: Union[list, tuple, np.ndarray],
+                 epsg: int) -> tuple:
+    """
+    Generates custom geo_info.
+    :param upper_left_xy: X and Y coordinates of upper left corner.
+    :param resol_xy: Resolution of image in X and Y directions.
+    :param epsg: EPSG
+    :return: transform and projection (i.e. geo_info).
+    """
+    transform = np.zeros(6)
+    transform[0] = upper_left_xy[0] - resol_xy[0] / 2
+    transform[1:3] = resol_xy[0], 0
+    transform[3] = upper_left_xy[1] + resol_xy[1] / 2
+    transform[4:] = 0, -resol_xy[1]
+
+    spatial_ref = osr.SpatialReference()
+    spatial_ref.ImportFromEPSG(epsg)
+    projection = spatial_ref.ExportToWkt()
+    geo_info = (transform, projection)
+    return geo_info
